-- LibSpecRoster
-- Project: LibSpecRoster r2
-- File: LibSpecRoster-1.0.lua
-- Last Modified: 2012-09-29T23:12:22Z
-- Author: msaint
-- Project desc: Efficiently maintains up to date raid spec information, inspecting
--               only when needed. 


--  Library methods:
--    lib.RegisterMessage(yourObject, "messageName", yourMethod)
--    lib.UnregisterMessage(yourObject, "messageName")
--    lib.RegisterAllMessages(yourObject)
--    lib:getSpecID(guid) - returns specialization ID
--    lib:getRole(guid) - returns role ("tank" | "healer" | "melee" | "ranged")
--    lib:getSpecInfo(guid) - returns specID, specName, class, role, blizzard role ("TANK" | "HEALER" | "DAMAGER") 
--    lib:getTalents(guid) - returns talents, oldTalents -- talents is an array.  Index is talent line (1-6), value is column of selected talent (1-3) 
--    lib:getGlyphs(guid) - returns majorGlyphs, minorGlyphs, oldMajorGlyphs, oldMinorGlyphs -- each is a table with glyphSpellID as keys
--
--
--  Messages:
--    "LSR_SpecializationChanged" - params: eventName, unit, guid, specID, oldSpecID
--    "LSR_RoleChanged" - params: eventName, unit, guid, role, oldRole
--    "LSR_TalentsChanged" - params: eventName, unit, guid, talentTable  -- talentTable is an array. Index is talent line (1-6), value is column of selected talent (1-3)
--    "LSR_GlyphsChanged" - params: eventName, unit, guid, major, minor  -- major and minor are tables with glyphSpellID as keys


-- Globals to ignore:
-- GLOBALS: DEFAULT_CHAT_FRAME, NUM_GLYPH_SLOTS, NotifyInspect

local DEBUG = false
local debug = DEBUG and function(s) DEFAULT_CHAT_FRAME:AddMessage("LSR: "..s, 1, 0, 0) end or function() return end  


-- ***** Obtain the library object **********
local LIB_VERSION = tonumber("2") or 5000
local MAJOR, MINOR = "LibSpecRoster-1.0", LIB_VERSION
local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end
local msg = LibStub("CallbackHandler-1.0"):New(lib, "RegisterMessage", "UnregisterMessage", "UnregisterAllMessages")


-- ***** LUA library functions made local **********
local math, max
      = math, math.max
local table, tinsert, tremove
      = table, tinsert, tremove
local pairs, ipairs, next, select, type, tostring
      = pairs, ipairs, next, select, type, tostring
local S = tostring

-- ***** API functions made local **********
local GetNumGroupMembers, GetPlayerInfoByGUID, IsInRaid, UnitGUID, UnitClass 
      = GetNumGroupMembers, GetPlayerInfoByGUID, IsInRaid, UnitGUID, UnitClass
local UnitIsVisible, UnitIsConnected, CanInspect, UnitIsUnit
      = UnitIsVisible, UnitIsConnected, CanInspect, UnitIsUnit
local GetSpecializationInfo, GetSpecialization, GetSpecializationInfoByID, GetInspectSpecialization
      = GetSpecializationInfo, GetSpecialization, GetSpecializationInfoByID, GetInspectSpecialization
local GetSpellInfo, GetTalentInfo, GetGlyphSocketInfo, GetClassInfo, GetNumClasses
      = GetSpellInfo, GetTalentInfo, GetGlyphSocketInfo, GetClassInfo, GetNumClasses
local RegisterEvent, UnregisterEvent, UnregisterAllEvents
      = RegisterEvent, UnregisterEvent, UnregisterAllEvents      


-- ***** Constants-ish **********
-- time delays
local updateDelay = 0.120 -- miliseconds between updates / talent requests
local notifyTimeout = 1 -- If requests don't come back in this time, requeue

-- Unit statuses
local GOOD, INVALID, QUEUED, REQUEUE, REQUESTED, REQUESTED_GOOD = "GOOD", "INVALID", "QUEUED", "REQUEUE", "REQUESTED", "REQUESTED_GOOD"
local valid_status = {GOOD = true, INVALID = true, QUEUED = true, REQUEUE = true, REQUESTED = true} -- REQUESTED_GOOD is not here because you cannot ask for it with setStatus

-- Class IDs, filled by init
local classID = {}

-- Role constants
local fixedRole = {
   ROGUE = "melee",
   HUNTER = "ranged",
   MAGE = "ranged",
   WARLOCK = "ranged",
}
local defaultDamageRole = {
   WARRIOR = "melee",
   DEATHKNIGHT = "melee",
   PALADIN = "melee",
   MONK = "melee",
}
setmetatable(defaultDamageRole,{ __index = function(l, class) return fixedRole[class] or "ranged" end})
local specIDRoles = {
      -- Death Knight
      [250] = "tank", -- Blood
      [251] = "melee", -- Frost
      [252] = "melee", -- Unholy
      -- Druid
      [102] = "ranged", -- Balance
      [103] = "melee", -- Feral
      [104] = "tank", -- Guardian
      [105] = "healer", -- Restoration
      -- Hunter
      [253] = "ranged", -- Beast Mastery
      [254] = "ranged", -- Marksmanship
      [255] = "ranged", -- Survival
      -- Mage
      [62] = "ranged", -- Arcane
      [63] = "ranged", -- Fire
      [64] = "ranged", -- Frost
      -- Monk
      [268] = "tank", -- Brewmaster
      [269] = "healer", -- Mistweaver
      [270] = "melee", -- Windwalker
      -- Paladin
      [65] = "healer", -- Holy
      [66] = "tank", -- Protection
      [67] = "melee", -- Retribution
      -- Priest
      [256] = "healer", -- Discipline
      [257] = "healer", -- Holy
      [258] = "ranged", -- Shadow
      -- Rogue
      [259] = "melee", -- Assassination
      [260] = "melee", -- Combat
      [261] = "melee", -- Subtlety
      -- Shaman
      [262] = "ranged", -- Elemental
      [263] = "melee", -- Enhancement
      [264] = "healer", -- Restoration
      -- Warlock
      [265] = "ranged", -- Affliction
      [266] = "ranged", -- Demonology
      [267] = "ranged", -- Destruction
      -- Warrior
      [71] = "melee", -- Arms
      [72] = "melee", -- Fury
      [73] = "tank", -- Protection
   }


-- ***** Local tables **********
local events = {} -- Used by event handlers and timers
local group = {} -- All raid members: group[guid] = {unitID, status, statusElapsed, class, specID, specName, oldSpecID, talents, oldTalents, glyphs, oldGlyphs}
                 -- group[].talents is an array of 6 numbers/nils representing the column of the selected talent at each tier. 
                 -- group[].glyphs = {major = {}, minor = {}} where major and minor are arrays of glyphs of that type.

-- ***** Local variables **********
local playerGUID -- Set in init() and used in setStatus()
--local houston = false -- if true, we are having problems getting class or guid info from the server


-- ***** Event Frame Setup **********
local eFrame = _G["LibSpecRosterEventFrame"] or CreateFrame("Frame", "LibSpecRosterEventFrame", UIParent)
eFrame:SetScript("OnEvent", function(self, event, ...)
      if events[event] then
         events[event](events, ...)
      end
   end) 
eFrame:RegisterEvent("PLAYER_ENTERING_WORLD") -- This is where we will initialize the library


-- ***** Local utility functions **********
local function shallowCopy(t)
   if t and type(t) == "table" then
      local t2 = {}
      for k, v in pairs(t) do
         t2[k] = v
      end
      return t2
   end
end

local function uniter() 
-- use is: for i, unitID in uniter() do block end
-- i starts at 1 and corresponds to raid or party unit number
-- for party (and solo), player is 'player' and is given last 
   local raid, N = IsInRaid(), GetNumGroupMembers() 
   N = N + ((N==0 or raid) and 1 or 0) 
   local iter = function(prefix, i)
      i = i + 1
      if (i < N) then
         return i, prefix..tostring(i)
      elseif (i == N and not raid) then
         return i, "player"
      else 
         return nil
      end
   end           
   return iter, (raid and "raid" or "party"), 0
end

local function buildRoster()
   local roster, guidRoster, guid = {}, {}
   for i, unit in uniter() do
      guid = UnitGUID(unit)
      if guid then
         roster[guid] = unit
      else
         debug("Houston, we need a procedure.  GUID was not available for "..S(unit))
      end
   end
   return roster
end

local function getRole(guid)
   debug("getRole received guid "..S(guid))
   local rec, role, isDefinite = group[guid]
   if rec then
      local class = rec.class or select(2, UnitClass(rec.unitID)) or "CLASSLESS"
      role = (rec.specID and (rec.specID > 0) and specIDRoles[rec.specID]) or defaultDamageRole[class]
      isDefinite = (fixedRole[class] or (rec.status == GOOD)) and true
      debug("getRole will return role of "..S(role).." for "..S(rec.unitID))
      return role, isDefinite
   end
end

local function getTalentTable(guid)
   local rec = group[guid]
   if rec then
      local unit = ((guid ~= playerGUID) and rec.unitID) or nil -- "player" needs no params except index
      local class = unit and ((rec.class and classID[rec.class]) or select(3, UnitClass(unit))) -- GetTalentInfo takes classID, not class string
      if (not unit) or class then -- Sometimes the client is still missing class info for other units, so abort
         local talents = {}
         for i=0, 5 do
            for j=1, 3 do
               if select(5, GetTalentInfo( i*3+j , unit, nil, unit, class)) then 
                  talents[i+1] = j
                  break
               end
            end
         end
         return (next(talents) and talents) or nil -- nil if no talents were found, it makes the result easier to use
      end
   end
end

local function getGlyphTable(guid) 
   local unit = ((guid ~= playerGUID) and group[guid].unitID) or nil -- "player" needs no params except index
   local major, minor = {}, {}
   for i=1, NUM_GLYPH_SLOTS do
      local _, rank, _, spellID = GetGlyphSocketInfo(i, nil, unit, unit)
      if spellID then
         major[spellID] = ((rank == 1) and true) or nil
         minor[spellID] = ((rank == 2) and true) or nil
      end
   end
   return ((next(major) or next(minor)) and {major = major, minor = minor}) or nil -- nil if no glyphs were found, it makes the result easier to use
end

local function fireMessages(guid)
   local rec = guid and group[guid]
   if rec then
      debug("Entered fireMessages for unit "..S(rec.unitID))
      if rec.talents then
         if rec.oldTalents then
            for line, col in ipairs(rec.talents) do
               if not rec.oldTalents[line] == col then
                  -- Fire talents changed
                  msg:Fire("LSR_TalentsChanged", rec.unitID, guid, rec.talents)
                  break
               end
            end
         else
            -- Also fire talents changed
            msg:Fire("LSR_TalentsChanged", rec.unitID, guid, rec.talents)
         end
      end
      if rec.glyphs then
         if rec.oldGlyphs then
            for glyph, _ in pairs(rec.glyphs) do
               if not rec.oldGlyphs[glyph] then
                  -- Fire glyphs changed
                  msg:Fire("LSR_GlyphsChanged", rec.unitID, guid, rec.glyphs)
                  break
               end
            end
         else
            -- Also fire glyphs changed
            msg:Fire("LSR_GlyphsChanged", rec.unitID, guid, rec.glyphs)
         end
      end
      if not (rec.specID == rec.oldSpecID) then 
         -- Fire spec changed
         msg:Fire("LSR_SpecializationChanged", rec.unitID, guid, rec.specID, rec.oldSpecID)
         if not (specIDRoles[rec.oldSpecID] == specIDRoles[rec.specID]) then
            -- Fire role changed
            msg:Fire("LSR_RoleChanged", rec.unitID, guid, specIDRoles[rec.specID], specIDRoles[rec.oldSpecID])
         end
      end
   end
end

local setStatus
do
   local lock 
   setStatus = function (guid, status)
      if not (guid and status and valid_status[status]) then return end
      while lock do end -- This is not perfect, but also not as bad as it looks
      lock = true
      local rec, doFire = group[guid]
      if rec then
         local oldStat = rec.status
         rec.statusElapsed = ((status == REQUESTED) and 0) or nil
         rec.status = ((status == REQUESTED) and (oldStat == GOOD) and REQUESTED_GOOD) or ((status == REQUEUE) and (oldStat == REQUESTED_GOOD) and GOOD)
                     or ( (((status == INVALID) and (oldStat == GOOD)) or ((status == GOOD) and ((guid == playerGUID) or (oldStat ~= INVALID)))
                        or ( (oldStat ~= GOOD) and ((status == QUEUED) or (status == REQUEUE) or (status == REQUESTED)) ) ) and status) or oldStat
      end
      lock = nil
      if (status == GOOD) and rec and (rec.status == GOOD) then
         fireMessages(guid)
      end
   end
end


-- I'm leaving the old version of setStatus here, because someone is going to tell me what an a*****e I am for
-- rewriting it that way.
--                     
-- local setStatus
-- do
--    local lock 
--    setStatus = function (guid, status)
--       if not (guid and status) then return end
--       while lock do end -- This is not perfect, but also not as bad as it looks
--       lock = true
--       local rec, doFire = group[guid]
--       if rec then
--          local oldStat = rec.status
--          if (status == GOOD) then
--             if (oldStat ~= INVALID) or (guid == playerGUID) then
--                -- 'player' info is instantly available, so if you set it to GOOD, I believe you.
--                -- Otherwise, don't jump from INVALID to GOOD, the data received may already be invalid.
--                rec.status = GOOD
--                rec.statusElapsed = nil
--                doFire = true
--             end
--          elseif (status == INVALID) then
--             if (oldStat == GOOD) then
--                rec.status = INVALID
--             end
--          elseif (status == REQUESTED) then
--             rec.status = (oldStat == GOOD) and REQUESTED_GOOD or REQUESTED
--             rec.statusElapsed = 0
--          elseif (status == REQUEUE) then
--             rec.status = ((oldStat == REQUESTED) and REQUEUE) or ((oldStat == REQUESTED_GOOD) and GOOD) or oldStat
--             rec.statusElapsed = nil
--          else
--             rec.status = status
--             rec.statusElapsed = nil
--          end
--       lock = nil
--       if rec then
--       end
--       if doFire then
--          fireMessages(guid)
--       end
--    end
-- end


-- ***** OnUpdate - Where it all happens! **********
local OnUpdate
do
   local sinceLast, queue = 0, {}, {}
   local checkOnce = true
   OnUpdate = function (self, elapsed)
      sinceLast = sinceLast + elapsed
      if (sinceLast >= updateDelay) then
         local bumpTime = sinceLast
         sinceLast = updateDelay - sinceLast -- REALLY do it every updateDelay ms
         
         for guid, record in pairs(group) do
            if (record.status == REQUESTED) or (record.status == REQUESTED_GOOD) then
               record.statusElapsed = record.statusElapsed + bumpTime
               if record.statusElapsed > notifyTimeout then
                  setStatus(guid, REQUEUE)
               end
            end       
            if (record.status == INVALID) or (record.status == REQUEUE) then
               tinsert(queue, 1, guid)
               setStatus(guid, QUEUED)
            end
         end
         
         for i = #queue, 1, -1 do
            local guid = queue[i] 
            local rec = group[guid]
            local unit = rec.unitID
            if (rec.status == QUEUED) and unit and CanInspect(unit) then
               if UnitIsConnected(unit) then
                  debug("Sending NotifyInspect for "..unit)
                  NotifyInspect(unit) -- REQUESTED status is set by our secure hook on NotifyInspect
                  tremove(queue, i)
                  break -- Only one request per requestThrottleDelay!
               elseif CanInspect(unit) then -- Always true for group members, even if they are offline
                  setStatus(guid, REQUEUE)
               end
            end
            tremove(queue, i)
         end
         
      end            
   end
end


-- ***** Library initialization **********
local function init()
   debug("init was reached")
   -- Fill classID table with className = ID
   for i=1, GetNumClasses() do
      local _, name, id = GetClassInfo(i) 
      classID[name] = id
   end   
   playerGUID = UnitGUID("player") -- Why look it up more times?
   -- Create initial roster before handling events
   events:GROUP_ROSTER_UPDATE()               
   eFrame:UnregisterAllEvents()
   for e, _ in pairs(events) do
      eFrame:RegisterEvent(e)
   end
   eFrame:Show()
end


function events:PLAYER_ENTERING_WORLD()
   debug("PLAYER_ENTERING_WORLD event handler was reached.")
   -- Ridiculous hack here just because I want to index by GUID
   local sinceLast, delay = 0, updateDelay
   local function doWaitForGUIDs(self, elapsed)
      sinceLast = sinceLast + elapsed
      if sinceLast >= delay then
         for _, unit in uniter() do
            if not UnitGUID(unit) then
               sinceLast = 0
               return
            end
         end
         eFrame:Hide()
         eFrame:SetScript("OnUpdate", OnUpdate)
         init()
      end 
   end
   eFrame:SetScript("OnUpdate", doWaitForGUIDs)
   eFrame:Show()
   events.PLAYER_ENTERING_WORLD = nil
end


-- ***** Event handlers **********
function events:GROUP_ROSTER_UPDATE()
   local newRoster = buildRoster()
   for guid, rec in pairs(group) do -- Discard or update existing entries in group
      if not newRoster[guid] then -- discard
         group[guid] = nil
      else -- same toon, new unitID: update and remove from newRoster
         rec.unitID = newRoster[guid]
         newRoster[guid] = nil
      end
   end 
   for guid, unit in pairs(newRoster) do -- Any units still in newRoster must be added to group 
      group[guid] = {unitID = unit, status = INVALID}
      group[guid].class = UnitClass(unit)
      if UnitIsUnit(unit, "player") then
         -- We should get this now: it shouldn't end up queued
         events:PLAYER_SPECIALIZATION_CHANGED("player")
      end
   end
end      


function events:PLAYER_SPECIALIZATION_CHANGED(unit)
   unit = unit or "player" -- This event sometimes leaves out unit for the player 
   debug("PLAYER_SPECIALIZATION_CHANGED event for "..S(unit))
   local guid = UnitGUID(unit)
   if guid then
      local record, newStatus = group[guid], INVALID
      if guid == playerGUID then
         local spec, specName, _, _, _, blizRole = GetSpecializationInfo((GetSpecialization())) 
         if spec and spec > 0 then
            record.oldTalents, record.oldGlyps, record.oldSpecID = record.talents, record.glyphs, record.specID
            record.specID, record.specName, record.blizRole = spec, specName, blizRole 
            record.class = record.class or UnitClass(unit)
            record.talents, record.glyphs = getTalentTable(guid), getGlyphTable(guid)
            newStatus = GOOD            
         end
      end
      setStatus(guid, newStatus)
   else
      debug("GUID was not returned for "..S(unit))
   end            
end


function events:INSPECT_READY(guid)
   debug("INSPECT_READY received for "..S(guid))
   local record = guid and group[guid]
   if record then
      local spec, specName, _, _, _, blizRole, class = GetSpecializationInfoByID(GetInspectSpecialization(record.unitID))
      if spec and spec > 0 then
         -- Even if the status doesn't become GOOD, this is the best data we have, so use it.
         record.oldTalents, record.oldGlyps, record.oldSpecID = record.talents, record.glyphs, record.specID
         record.specID, record.specName, record.blizRole, record.class = spec, specName, blizRole, class
         record.talents, record.glyphs = getTalentTable(guid), getGlyphTable(guid)         
         setStatus(guid, GOOD)
      end
   end
end


-- ***** Hook to catch inspects from other addons and avoid duplicating them **********
local function notifyInspect(unit)
   local guid = UnitGUID(unit)
   if (guid and group[guid] and CanInspect(unit) and UnitIsConnected(unit)) then
	   -- If we know the request is going to fail, don't flag it as REQUESTED
      setStatus(guid, REQUESTED)
	end
end
hooksecurefunc("NotifyInspect", notifyInspect)


-- ***** Exposed Library Methods **********
--
function lib:getSpecID(guid)
   local rec = guid and group[guid]
   if rec then
      return rec.specID, ((rec.status == GOOD) or (rec.status == REQUESTED_GOOD))
   end
end

function lib:getROle(guid)
   if guid and group[guid] then
      return getRole(guid)
   end
end

function lib:getSpecInfo(guid)
   local rec = guid and group[guid]
   if rec then
      return rec.specID, rec.specName, (rec.class or (select(2, UnitClass(rec.unitID)))), (getRole(guid)), rec.blizRole 
   end
end

function lib:getTalents(guid)
   local rec = guid and group[guid]
   if rec and rec.talents then
      return shallowCopy(rec.talents), (rec.oldTalents and shallowCopy(rec.oldTalents))
   end
end

function lib:getGlyphs(guid)
   local rec = guid and group[guid]
   if rec and rec.glyphs then
      return (rec.glyphs.major and shallowCopy(rec.glyphs.major)), (rec.glyphs.minor and shallowCopy(rec.glyphs.minor)),
             (rec.oldGlyphs.major and shallowCopy(rec.oldGlyphs.major)), (rec.oldGlyphs.minor and shallowCopy(rec.oldGlyphs.minor))
   end
end
